from servicelayer import settings
from servicelayer.archive.file import FileArchive

ARCHIVE_FILE = "file"
ARCHIVE_S3 = "s3"
ARCHIVE_GS = "gs"


def init_archive(
    archive_type=settings.ARCHIVE_TYPE,
    path=settings.ARCHIVE_PATH,
    bucket=settings.ARCHIVE_BUCKET,
    publication_bucket=settings.PUBLICATION_BUCKET,
):
    """Instantiate an archive object."""
    if archive_type == ARCHIVE_S3:
        from servicelayer.archive.s3 import S3Archive

        return S3Archive(bucket=bucket, publication_bucket=publication_bucket)

    if archive_type == ARCHIVE_GS:
        from servicelayer.archive.gs import GoogleStorageArchive

        return GoogleStorageArchive(
            bucket=bucket, publication_bucket=publication_bucket
        )

    return FileArchive(path=path)



import abc


class Archive(object):
    __metaclass__ = abc.ABCMeta

    def upgrade(self):
        """Run maintenance on the store."""
        pass

    @abc.abstractmethod
    def archive_file(self, file_path, content_hash=None, mime_type=None):
        """Import the given file into the archive."""
        pass

    @abc.abstractmethod
    def load_file(self, content_hash, file_name=None, temp_path=None):
        pass

    @abc.abstractmethod
    def list_files(self, prefix=None):
        """List files in the archive within the given prefix. Must return an
        iterator of `content_hash`."""
        pass

    @abc.abstractmethod
    def delete_file(self, content_hash):
        pass

    def cleanup_file(self, content_hash, temp_path=None):
        pass

    def generate_url(self, content_hash, file_name=None, mime_type=None, expire=None):
        return None

    @property
    def can_publish(self):
        return False

    def publish_file(self, file_path, publish_path, mime_type=None):
        return None



import time
import random
import logging
import uuid
from banal import ensure_list

from redis.exceptions import BusyLoadingError

from servicelayer.settings import REDIS_EXPIRE
from servicelayer.settings import REDIS_PREFIX as PREFIX
from servicelayer.cache import make_key
from servicelayer.util import pack_now, dump_json, load_json
from servicelayer.util import unpack_int, sum_values

log = logging.getLogger(__name__)


class Dataset(object):
    """A subdivision of tasks. Each dataset can have multiple stages
    of processing and jobs that control processing instances."""

    def __init__(self, conn, name):
        self.conn = conn
        self.name = name
        self.key = make_key(PREFIX, "qdatasets")
        self.jobs_key = make_key(PREFIX, "qdj", name)
        self.stages_key = make_key(PREFIX, "qds", name)

    def cancel(self):
        pipe = self.conn.pipeline()
        for job in self.get_jobs():
            job._remove(pipe)
        pipe.delete(self.stages_key)
        pipe.delete(self.jobs_key)
        pipe.execute()

    def get_stages(self):
        return self.conn.smembers(self.stages_key)

    def get_job_ids(self):
        return self.conn.smembers(self.jobs_key)

    def get_jobs(self):
        for job_id in self.get_job_ids():
            yield Job(self.conn, self, job_id)

    def get_status(self):
        """Aggregate status for all stages on the given dataset."""
        status = {"finished": 0, "running": 0, "pending": 0, "jobs": []}
        for job in self.get_jobs():
            progress = job.get_status()
            status["jobs"].append(progress)
            status["finished"] += progress["finished"]
            status["running"] += progress["running"]
            status["pending"] += progress["pending"]
        return status

    def __str__(self):
        return self.name

    @classmethod
    def ensure(cls, conn, name):
        if isinstance(name, cls):
            return name
        return cls(conn, name)

    @classmethod
    def get_active_datasets(cls, conn):
        datasets_key = make_key(PREFIX, "qdatasets")
        for name in conn.smembers(datasets_key):
            yield cls(conn, name)

    @classmethod
    def get_active_dataset_status(cls, conn):
        result = {"total": 0, "datasets": {}}
        for dataset in cls.get_active_datasets(conn):
            status = dataset.get_status()
            result["total"] += 1
            result["datasets"][dataset.name] = status
        return result


class Job(object):
    def __init__(self, conn, dataset, job_id):  # noqa
        self.conn = conn
        self.id = job_id
        self.dataset = Dataset.ensure(conn, dataset)
        self.start_key = make_key(PREFIX, "qd", self.id, dataset, "start")
        self.end_key = make_key(PREFIX, "qd", self.id, dataset, "end")
        self.last_update_key = make_key(PREFIX, "qd", self.id, dataset, "last_update")
        self.active_jobs_key = make_key(PREFIX, "qdja")

    def get_stage(self, name):
        return Stage(self, name)

    def get_stages(self):
        for stage in self.dataset.get_stages():
            yield self.get_stage(stage)

    def is_done(self):
        if self.conn.exists(self.end_key):
            return True
        for _ in range(5):
            keys = self._get_active_keys()
            active = sum_values(self.conn.mget(keys))
            if active > 0:
                return False

            for stage in self.get_stages():
                pending = self.conn.llen(stage.queue_key)
                self.conn.set(stage.pending_key, pending)
        self.conn.setnx(self.end_key, pack_now())
        return True

    def _create(self, pipe):
        pipe.sadd(self.dataset.key, self.dataset.name)
        pipe.sadd(self.dataset.jobs_key, self.id)
        pipe.sadd(self.active_jobs_key, make_key(self.dataset.name, self.id))
        pipe.delete(self.end_key)
        pipe.setnx(self.start_key, pack_now())
        self.touch_last_updated(pipe)

    def _remove(self, pipe):
        for stage in self.get_stages():
            stage._remove(pipe)
        pipe.srem(self.dataset.key, self.dataset.name)
        pipe.srem(self.dataset.jobs_key, self.id)
        pipe.srem(self.active_jobs_key, make_key(self.dataset.name, self.id))
        pipe.delete(self.start_key)
        pipe.delete(self.last_update_key)
        pipe.setnx(self.end_key, pack_now())
        pipe.expire(self.end_key, REDIS_EXPIRE)

    def remove(self):
        pipe = self.conn.pipeline()
        self._remove(pipe)
        pipe.execute()

    def _get_active_keys(self):
        """Return the pending keys for all stages in this job"""
        keys = []
        for stage in self.get_stages():
            keys.append(stage.pending_key)
            keys.append(stage.running_key)
        return keys

    def get_status(self):
        """Aggregate status for all stages on the given job."""
        status = {"finished": 0, "running": 0, "pending": 0, "stages": []}
        start, end, last_update = self.conn.mget(
            (self.start_key, self.end_key, self.last_update_key)
        )
        status["start_time"] = start
        status["end_time"] = end
        status["last_update"] = last_update
        for stage in self.get_stages():
            progress = stage.get_status()
            status["stages"].append(progress)
            status["finished"] += progress["finished"]
            status["running"] += progress["running"]
            status["pending"] += progress["pending"]
        return status

    def touch_last_updated(self, pipe):
        """Update the last_updated key for this job."""
        pipe.set(self.last_update_key, pack_now())

    @classmethod
    def random_id(cls):
        return uuid.uuid4().hex

    @classmethod
    def create(cls, conn, dataset):
        return cls(conn, dataset=dataset, job_id=Job.random_id())


class Stage(object):
    def __init__(self, job, stage):  # noqa
        self.job = job
        self.conn = job.conn
        self.stage = stage
        self.stages_key = self._get_stage_jobs_key(stage)
        queue_id = (PREFIX, "q", job.dataset, stage, job.id)
        self.queue_key = make_key(*queue_id)
        self.pending_key = make_key(*queue_id, "pending")
        self.running_key = make_key(*queue_id, "running")
        self.finished_key = make_key(*queue_id, "finished")

    def _create(self, pipe):
        pipe.sadd(self.stages_key, self.queue_key)
        pipe.sadd(self.job.dataset.stages_key, self.stage)
        self.job._create(pipe)

    def _remove(self, pipe):
        """Remove tasks for the current `job_id` and `stage`"""
        pipe.srem(self.stages_key, self.queue_key)
        pipe.delete(
            self.queue_key, self.pending_key, self.running_key, self.finished_key
        )

    def _check_out(self, count=1):
        """Check out tasks and mark them as running."""
        pipe = self.conn.pipeline()
        self._create(pipe)
        pipe.decr(self.pending_key, amount=count)
        pipe.incr(self.running_key, amount=count)
        self.job.touch_last_updated(pipe)
        pipe.execute()

    def mark_done(self, count=1):
        """Returns tasks previously checked out and mark as done."""
        pipe = self.conn.pipeline()
        self._create(pipe)
        pipe.decr(self.running_key, amount=count)
        pipe.incr(self.finished_key, amount=count)
        self.job.touch_last_updated(pipe)
        pipe.execute()

    def report_finished(self, count=1):
        """Inflate the number of finished tasks."""
        pipe = self.conn.pipeline()
        self._create(pipe)
        pipe.incr(self.finished_key, amount=count)
        self.job.touch_last_updated(pipe)
        pipe.execute()

    def queue(self, payload={}, context={}):
        task = Task(self, payload, context)
        data = task.serialize()
        pipe = self.conn.pipeline()
        self._create(pipe)
        pipe.rpush(self.queue_key, data)
        pipe.incr(self.pending_key)
        self.job.touch_last_updated(pipe)
        pipe.execute()
        return task

    def sync(self):
        pending = self.conn.llen(self.queue_key)
        self.conn.set(self.pending_key, pending)

    def get_tasks(self, limit=100):
        """Get multiple tasks at once, without blocking. This is used
        inside the consumer applications to process multiple tasks of
        the same type at once."""
        if limit is None or limit < 1:
            return []
        pipe = self.conn.pipeline()
        pipe.lrange(self.queue_key, 0, limit - 1)
        pipe.ltrim(self.queue_key, limit, -1)
        raw_tasks = pipe.execute()[0]
        tasks = []
        for task in raw_tasks:
            task = Task.unpack(self.conn, task)
            tasks.append(task)
        # TODO: can this be atomic?
        self._check_out(len(tasks))
        return tasks

    def get_status(self):
        """Get the current status."""
        keys = (self.pending_key, self.running_key, self.finished_key)
        pending, running, finished = self.conn.mget(keys)
        return {
            "job_id": self.job.id,
            "stage": self.stage,
            "pending": max(0, unpack_int(pending)),
            "running": max(0, unpack_int(running)),
            "finished": max(0, unpack_int(finished)),
        }

    @classmethod
    def _get_stage_jobs_key(cls, stage):
        return make_key(PREFIX, "qos", stage)

    @classmethod
    def _get_queues(cls, conn, stages):
        """Return all the active queues for the given stage."""
        queues = []
        for stage in ensure_list(stages):
            key = cls._get_stage_jobs_key(stage)
            queues.extend(conn.smembers(key))
        # TODO: do we want to random.shuffle?
        random.shuffle(queues)
        return queues

    @classmethod
    def get_task(cls, conn, stages, timeout=0):
        """Retrieve a single task from the highest-priority queue that has
        work pending."""
        try:
            queues = cls._get_queues(conn, stages)
            if not len(queues):
                # Avoid going into a tight loop when there are no tasks:
                # cf. https://github.com/alephdata/aleph/issues/867
                if timeout is not None:
                    time.sleep(max(10, timeout or 1))
                return None
            # Support a magic value to not block, i.e. timeout None
            if timeout is None:
                # LPOP does not support multiple lists.
                for queue in queues:
                    task_data = conn.lpop(queue)
                    if task_data is not None:
                        break
                if task_data is None:
                    return None
            else:
                task_data = conn.blpop(queues, timeout=timeout)
                if task_data is None:
                    return None
                _, task_data = task_data

            task = Task.unpack(conn, task_data)
            # TODO: can this be atomic?
            task.stage._check_out(1)
            return task
        except BusyLoadingError:
            time.sleep(timeout + 1)
            return None


class Task(object):
    def __init__(self, stage, payload, context):
        self.payload = payload
        self.context = context
        self.stage = stage
        self.job = stage.job

    def done(self):
        self.stage.mark_done(1)

    def serialize(self):
        return dump_json(
            {
                "context": self.context or {},
                "payload": self.payload,
                "dataset": self.job.dataset.name,
                "job": self.job.id,
                "stage": self.stage.stage,
            }
        )

    @classmethod
    def unpack(cls, conn, data):
        if data is None:
            return None
        data = load_json(data)
        job = Job(conn, data.get("dataset"), data.get("job"))
        stage = job.get_stage(data.get("stage"))
        return Task(stage, data.get("payload"), data.get("context"))



from datetime import datetime
from normality import stringify

from servicelayer.jobs import Job
from servicelayer.settings import WORKER_REPORTING


OP_REPORT = "report"


class Status:
    START = "start"
    SUCCESS = "success"
    ERROR = "error"


class Reporter(object):
    """A reporter will create processing status updates."""

    def __init__(self, task=None, stage=None, operation=None):
        self.task = task
        self.stage = stage
        self.operation = operation

    def start(self, **data):
        self.handle(status=Status.START, **data)

    def end(self, **data):
        self.handle(status=Status.SUCCESS, **data)

    def error(self, exception, **data):
        self.handle(status=Status.ERROR, exception=exception, **data)

    def handle(self, status, operation=None, exception=None, task=None, **payload):
        """Report a processing event that may be related to a task."""
        if not WORKER_REPORTING:
            return

        task = task or self.task
        if task is not None:
            payload["task"] = task.serialize()
            stage = task.stage
        else:
            stage = self.stage
        dataset = stage.job.dataset.name
        job_id = stage.job.id
        operation = operation or stage.stage

        now = datetime.utcnow()
        payload.update(
            {
                "dataset": dataset,
                "operation": operation,
                "job": job_id,
                "status": status,
                "updated_at": now,
                "%s_at" % status: now,
                "has_error": False,
            }
        )

        if exception is not None:
            payload.update(
                {
                    "status": Status.ERROR,
                    "has_error": True,
                    "error_name": exception.__class__.__name__,
                    "error_msg": stringify(exception),
                }
            )

        job = Job(stage.conn, dataset, job_id)
        stage = job.get_stage(OP_REPORT)
        stage.queue(payload)



from unittest import TestCase


from servicelayer.cache import get_fakeredis
from servicelayer.taskqueue import (
    Dataset,
    dataset_from_collection_id,
)


class TestDataset(TestCase):
    def setUp(self):
        self.connection = get_fakeredis()
        self.connection.flushdb()
        self.collection_id = 1

        self.dataset = Dataset(
            conn=self.connection, name=dataset_from_collection_id(self.collection_id)
        )

    def test_get_active_datasets_key(self):
        assert self.dataset.key == "tq:qdatasets"

    def test_get_active_stages_key(self):
        assert (
            self.dataset.active_stages_key
            == f"tq:qds:{self.collection_id}:active_stages"
        )

    def test_get_timestamp_keys(self):
        assert self.dataset.start_key == f"tq:qdj:{self.collection_id}:start"
        assert (
            self.dataset.last_update_key == f"tq:qdj:{self.collection_id}:last_update"
        )

    def test_tasks_per_collection_keys(self):
        assert self.dataset.finished_key == f"tq:qdj:{self.collection_id}:finished"
        assert self.dataset.running_key == f"tq:qdj:{self.collection_id}:running"
        assert self.dataset.pending_key == f"tq:qdj:{self.collection_id}:pending"



import logging
import pytest
from prometheus_client import REGISTRY
from prometheus_client.metrics import MetricWrapperBase

from servicelayer.cache import get_fakeredis
from servicelayer.jobs import Job
from servicelayer import worker


worker.TASK_FETCH_RETRY = 1


class CountingWorker(worker.Worker):
    def boot(self):
        self.test_done = 0

    def handle(self, task):
        self.test_done += 1


class FailingWorker(worker.Worker):
    def handle(self, task):
        raise Exception("Woops")


class NoOpWorker(worker.Worker):
    def handle(self, task):
        pass


class PrometheusTests:
    def setup_method(self, method):
        # This relies on internal implementation details of the client to reset
        # previously collected metrics before every test execution. Unfortunately,
        # there is no clean way of achieving the same thing that doesn't add a lot
        # of complexity to the test and application code.
        collectors = REGISTRY._collector_to_names.keys()
        for collector in collectors:
            if isinstance(collector, MetricWrapperBase):
                collector._metrics.clear()
                collector._metric_init()

    def test_prometheus_succeeded(self):
        conn = get_fakeredis()
        worker = CountingWorker(conn=conn, stages=["ingest"])
        job = Job.create(conn, "test")
        stage = job.get_stage("ingest")
        stage.queue({}, {})
        worker.sync()

        labels = {"stage": "ingest"}
        success_labels = {"stage": "ingest", "retries": "0"}

        started = REGISTRY.get_sample_value("task_started_total", labels)
        succeeded = REGISTRY.get_sample_value("task_succeeded_total", success_labels)

        # Under the hood, histogram metrics create multiple time series tracking
        # the number and sum of observations, as well as individual histogram buckets.
        duration_sum = REGISTRY.get_sample_value("task_duration_seconds_sum", labels)
        duration_count = REGISTRY.get_sample_value(
            "task_duration_seconds_count",
            labels,
        )

        assert started == 1
        assert succeeded == 1
        assert duration_sum > 0
        assert duration_count == 1

    def test_prometheus_failed(self):
        conn = get_fakeredis()
        worker = FailingWorker(conn=conn, stages=["ingest"])
        job = Job.create(conn, "test")
        stage = job.get_stage("ingest")
        stage.queue({}, {})
        labels = {"stage": "ingest"}

        worker.sync()

        assert REGISTRY.get_sample_value("task_started_total", labels) == 1
        assert REGISTRY.get_sample_value(
            "task_failed_total",
            {
                "stage": "ingest",
                "retries": "0",
                "failed_permanently": "False",
            },
        )

        worker.sync()

        assert REGISTRY.get_sample_value("task_started_total", labels) == 2
        assert REGISTRY.get_sample_value(
            "task_failed_total",
            {
                "stage": "ingest",
                "retries": "1",
                "failed_permanently": "False",
            },
        )

        worker.sync()

        assert REGISTRY.get_sample_value("task_started_total", labels) == 3
        assert REGISTRY.get_sample_value(
            "task_failed_total",
            {
                "stage": "ingest",
                "retries": "2",
                "failed_permanently": "False",
            },
        )

        worker.sync()

        assert REGISTRY.get_sample_value("task_started_total", labels) == 4
        assert REGISTRY.get_sample_value(
            "task_failed_total",
            {
                "stage": "ingest",
                "retries": "3",
                "failed_permanently": "True",
            },
        )


def test_run():
    conn = get_fakeredis()
    operation = "lala"
    worker = CountingWorker(conn=conn, stages=[operation])
    worker.sync()
    assert worker.test_done == 0, worker.test_done
    job = Job.create(conn, "test")
    stage = job.get_stage(operation)
    task = stage.queue({}, {})
    assert not job.is_done()
    assert worker.test_done == 0, worker.test_done
    worker.sync()
    assert worker.test_done == 1, worker.test_done
    assert job.is_done()
    worker.retry(task)
    assert not job.is_done()
    worker.sync()
    assert job.is_done()
    assert worker.exit_code == 0, worker.exit_code
    assert worker.test_done == 1, worker.test_done
    worker.retry(task)
    worker.run(blocking=False)
    assert job.is_done()
    assert worker.exit_code == 0, worker.exit_code
    worker.num_threads = None
    worker.retry(task)
    worker.run(blocking=False)
    assert job.is_done()
    assert worker.exit_code == 0, worker.exit_code
    try:
        worker._handle_signal(5, None)
    except SystemExit as exc:
        assert exc.code == 5, exc.code
    with pytest.raises(SystemExit) as exc:  # noqa
        worker._handle_signal(5, None)


def test_fails(caplog):
    caplog.set_level(logging.DEBUG)
    conn = get_fakeredis()
    operation = "fails"
    job = Job.create(conn, "test")
    stage = job.get_stage(operation)
    task = stage.queue({}, {})

    worker = NoOpWorker(conn=conn, stages=[operation])
    worker.sync()
    for _ in range(4):
        worker.retry(task)

    log_messages = [r.msg for r in caplog.records]
    assert "Queueing failed task for retry #1/3..." in log_messages
    assert "Queueing failed task for retry #2/3..." in log_messages
    assert "Queueing failed task for retry #3/3..." in log_messages
    assert "Failed task, exhausted retry count of 3" in log_messages
